#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Class for polling Ceilometer

This class provides means to requests for authentication
tokens to be used with OpenStack's Ceilometer, Nova and RabbitMQ
"""

from eszcp.ceilometer_client import Client as CeiloV20
from eszcp.common import log
from eszcp import utils
from eszcp.task.polling import INSTANCE_METRICS, NETWORK_METRICS
from eszcp.task.polling.base_handler import Handler


LOG = log.logger(__name__)

METRIC_CACEHES = {}


def get_handler(*args):
    return CeilometerHandler(*args)


class CeilometerHandler(Handler):

    def __init__(self, conf,
                 ks_client, nv_client, zabbix_hdl):
        """Ceilometer polling handler

        :param conf: zcp configs
        :parms ks_client: Openstack identity service, keystone client
        :parms nv_client: Openstack compute service,nova client
        :parms zabbix_hdl:zabbix handler
        """
        super(CeilometerHandler, self).__init__(conf, zabbix_hdl)
        self.polling_interval = int(conf.read_option('zcp_configs',
                                                     'polling_interval'))
        self.ks_client = ks_client
        self.nv_client = nv_client
        self.zabbix_hdl = zabbix_hdl
        self.clt_client = CeiloV20()

    def polling_metrics(self, hosts, hosts_map):
        """
        For ES metric collecor,
        We don't storage sample in a same collection meter(
        OpenStack community do it like this).

        Accroding to the sample period,slice the meter collection
        and collector metric
        :param hosts: a list ,record zabbix host information.
        :param hosts_cache:a dict ,record zabbix host information.
         hosts_cache, the date structure is the following:
         {"instance_id":["host_id","host_name","proxy_id"]
         }
        hosts: ['instance_id1','instance)id2',...]
        """
        def _all_instance_details():
            response = self.nv_client.instance_get_all()
            servers = [server.to_dict() for server in response]
            if not servers:
                LOG.warning("Servers list is empry,"
                            "skip to update zabbix values")
            return servers

        All_INSTANCES = _all_instance_details() or []
        for instance in All_INSTANCES:
            if instance['id'] in hosts and utils.is_active(instance):
                LOG.debug("Start Checking host : %s"
                          % hosts_map[instance['id']][1])
                # Get links for instance compute metrics
                query = [{'op': 'eq',
                          'value': instance['id'],
                          'field': 'metadata.instance_id'}]
                resources = self.clt_client.list_resources(q=query)
                # Add a new instance and its metrics
                if instance['id'] not in METRIC_CACEHES.keys():
                    rs_items = {}
                    for rs in resources:
                        if rs.resource_id.startswith('instance'):
                            rs_items[rs.resource_id] = NETWORK_METRICS
                        # NOTE:remove disk metrics
                        elif utils.endswith_words(rs.resource_id):
                            pass
                        else:
                            rs_items[rs.resource_id] = INSTANCE_METRICS
                    METRIC_CACEHES[instance['id']] = rs_items
                # Update metric_caches where instance_in exists.For the case:
                # instance add/remove a nic
                # instance add/remove a volume
                else:
                    rs_items = METRIC_CACEHES[instance['id']]
                    rs_item_keys = rs_items.keys()
                    for rs in resources:
                        if rs.resource_id not in rs_item_keys and \
                           rs.resource_id.startswith('instance'):
                            rs_items[rs.resource_id] = NETWORK_METRICS
                            METRIC_CACEHES[instance['id']] = rs_items
                        # NOTE:remove disk metrics
                        elif rs.resource_id not in rs_item_keys and \
                                utils.endswith_words(rs.resource_id):
                            pass
                        else:
                            continue
                LOG.debug("Starting to polling %s(%s) metric into zabbix"
                          % (instance.get('name'), instance.get('id')))
                pxy = self.zabbix_hdl.get_by_proxyid(
                                hosts_map[instance['id']][2])
                if pxy:
                    proxy_name = pxy['host']
                else:
                    LOG.warning("Can't find the prxoy:%s,Skip to polling "
                                "instance_id  %s metrics."
                                % (hosts_map[instance['id']][2],
                                   instance['id']))
                    continue
                # Polling Ceilometer the latest sample into zabbix
                # CLI:ceilometer statistics -m {...} -q resource_id={...} -p ..
                self._polling_metrics(instance['id'],
                                      proxy_name)
                LOG.debug("Finshed to polling %s(%s) metric into zabbix"
                          % (instance.get('name'), instance.get('id')))
            else:
                LOG.debug("Can't find the instance : %s(%s), "
                          "or the status of %s is not active"
                          % (instance.get('name'),
                             instance.get('id'),
                             instance.get('name'))
                          )

    def _polling_metrics(self, instance_id, proxy_name):
        """
        :param instance_id: nova instance uuid
        :param proxy_name: zabbix proxy_name
        """
        def _polling(ids, METRICS):
            for metric in METRICS:
                counter_volume = 0.0
                for rsc_id in ids:
                    query = [{'op': 'eq',
                              'value': rsc_id,
                              'field': 'resource_id'}]
                    response = self.clt_client.statistics(
                                metric,
                                q=query,
                                limit=1)
                    if len(response) > 0:
                        counter_volume += response[0].avg
                    else:
                        LOG.info("The metric %s of resource_id %s statistics "
                                 "not found." % (metric, rsc_id))
                    LOG.info("Polling Ceilometer metri into zabbix proxy: %s,"
                             "resource_id: %s, metric: %s, counter_name: %s"
                             % (rsc_id, metric, counter_volume, proxy_name))
                    self.zabbix_hdl.send_data_zabbix(counter_volume,
                                                     instance_id,
                                                     metric,
                                                     proxy_name)
        # Get instance all taps
        network_nics_id = []
        # Get install all volumes
        for rsc_id in METRIC_CACEHES[instance_id].keys():
            if rsc_id.startswith('instance'):
                network_nics_id.append(rsc_id)

        # network metrics
        _polling(network_nics_id, NETWORK_METRICS)

        # instance metrics
        _polling([instance_id], INSTANCE_METRICS)



#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
from novaclient import client as nova_client
from novaclient import exceptions

from eszcp.common import log


LOG = log.logger(__name__)


def logged(func):

    @functools.wraps(func)
    def with_logging(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exceptions.Unauthorized, e:
            msg = "Error... \nToken refused! " \
                  "The request you have made requires authentication."
            LOG.error(msg)
            raise
        except exceptions.NotFound, e:
            LOG.error("Not Found Nova Resource")
            raise
        except Exception, ex:
            msg = getattr(ex, 'message', None) or \
                  getattr(ex, 'msg', '')
            LOG.error(msg)
            raise
    return with_logging


class Client(object):
    def __init__(self, conf):
        # novaclient only support keystone v3
        auth_url = conf.read_option('keystone_authtoken',
                                    'auth_url')
        if auth_url.endswith('/v3'):
            auth_url = auth_url.replace('/v3', '/v2.0')
        self.nv_client = nova_client.Client(2,
                                            conf.read_option(
                                                'keystone_authtoken',
                                                'username'),
                                            conf.read_option(
                                                'keystone_authtoken',
                                                'password'),
                                            conf.read_option(
                                                'keystone_authtoken',
                                                'project_name'),
                                            auth_url,
                                            region_name=conf.read_option(
                                                'keystone_authtoken',
                                                'region_name')
                                            )

    @logged
    def instance_get_all(self, since=None):
        """Returns list of all instances.

        If since is supplied, it will return the instances changes since that
        datetime. since should be in ISO Format '%Y-%m-%dT%H:%M:%SZ'
        """
        search_opts = {'all_tenants': True}
        if since:
            search_opts['changes-since'] = since
        return self.nv_client.servers.list(
            detailed=True,
            search_opts=search_opts)



#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Class for requesting authentication tokens to Keystone

This class provides means to requests for authentication

tokens to be used with OpenStack's Ceilometer, Nova and RabbitMQ

and query requirements for keystone projects and domains
"""

import functools
from keystoneclient.v3 import client as ks_client_v3

from eszcp.common import log


LOG = log.logger(__name__)


def logged(func):

    @functools.wraps(func)
    def with_logging(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            LOG.error(e)
            raise e
    return with_logging


class Client(object):
    def __init__(self, conf):
        v3_kwargs = {
                "username": conf.read_option('keystone_authtoken',
                                             'username'),
                "password": conf.read_option('keystone_authtoken',
                                             'password'),
                "project_name": conf.read_option(
                                             'keystone_authtoken',
                                             'project_name'),
                "user_domain_name": conf.read_option(
                                             'keystone_authtoken',
                                             'user_domain_name'),
                "project_domain_name": conf.read_option(
                                             'keystone_authtoken',
                                             'project_domain_name'),
                "auth_url": conf.read_option('keystone_authtoken',
                                             'auth_url'),
                "region_name": conf.read_option(
                                             'keystone_authtoken',
                                             'region_name'),
        }
        # project scope keystoneclient
        self.project_keystone = ks_client_v3.Client(**v3_kwargs)
        del v3_kwargs['project_name']
        v3_kwargs['domain_name'] = conf.read_option('keystone_authtoken',
                                                    'domain_name'
                                                    )
        # domain scope keystoneclient
        self.domain_keystone = ks_client_v3.Client(**v3_kwargs)

    @logged
    def get_domains(self):
        return self.domain_keystone.domains.list()

    @logged
    def show_domain(self, domain_id_or_name):
        return self.domain_keystone.domains.get(domain_id_or_name)

    @logged
    def get_projects(self, domain_id=None):
        if domain_id:
            return self.domain_keystone.projects.list(domain=domain_id)
        else:
            return self.domain_keystone.projects.list()

    @logged
    def get_project(self, project_id):
        return self.domain_keystone.projects.get(project_id)



#    Copyright  2017 EasyStack, Inc
#    Authors: Hanxi Liu<apolloliuhx@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exception definitions."""


class ZcpException(Exception):
    """The base exception class for all exceptions"""
    pass


class TemplateNotFound(ZcpException):
    """"Not find default template in zabbix"""
    pass


class NotImplementedError(NotImplementedError):
    """Not Implementation"""
    pass


class MappingFileNotFound(ZcpException):
    """Not find mapping.json"""
    pass



#    Copyright  2017 EasyStack, Inc
#    Author : Branty(jun.wang@easystack.cn)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Pymongo Utilities and helper functions."""

import six
from six.moves.urllib import parse


def quote_key(key, reverse=False):
    """Prepare key for storage data in MongoDB.

    :param key: key that should be quoted
    :param reverse: boolean, True --- if we need a reverse order of the keys
                    parts
    :return: iter of quoted part of the key
    """
    r = -1 if reverse else 1

    for k in key.split('.')[::r]:
        if k.startswith('$'):
            k = parse.quote(k)
        yield k


def improve_keys(data, metaquery=False):
    """Improves keys in dict if they contained '.' or started with '$'.

    :param data: is a dictionary where keys need to be checked and improved
    :param metaquery: boolean, if True dots are not escaped from the keys
    :return: improved dictionary if keys contained dots or started with '$':
            {'a.b': 'v'} -> {'a': {'b': 'v'}}
            {'$ab': 'v'} -> {'%24ab': 'v'}
    """
    if not isinstance(data, dict):
        return data

    if metaquery:
        for key in six.iterkeys(data):
            if '.$' in key:
                key_list = []
                for k in quote_key(key):
                    key_list.append(k)
                new_key = '.'.join(key_list)
                data[new_key] = data.pop(key)
    else:
        for key, value in data.items():
            if isinstance(value, dict):
                improve_keys(value)
            if '.' in key:
                new_dict = {}
                for k in quote_key(key, reverse=True):
                    new = {}
                    new[k] = new_dict if new_dict else data.pop(key)
                    new_dict = new
                data.update(new_dict)
            else:
                if key.startswith('$'):
                    new_key = parse.quote(key)
                    data[new_key] = data.pop(key)
    return data


def unquote_keys(data):
    """Restores initial view of 'quoted' keys in dictionary data

    :param data: is a dictionary
    :return: data with restored keys if they were 'quoted'.
    """
    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, dict):
                unquote_keys(value)
            if key.startswith('%24'):
                k = parse.unquote(key)
                data[k] = data.pop(key)
    return data


def make_timestamp_range(start, end,
                         start_timestamp_op=None, end_timestamp_op=None):

    """Create the query document to find timestamps within that range.

    This is done by given two possible datetimes and their operations.
    By default, using $gte for the lower bound and $lt for the upper bound.
    """
    ts_range = {}

    if start:
        if start_timestamp_op == 'gt':
            start_timestamp_op = '$gt'
        else:
            start_timestamp_op = '$gte'
        ts_range[start_timestamp_op] = start

    if end:
        if end_timestamp_op == 'le':
            end_timestamp_op = '$lte'
        else:
            end_timestamp_op = '$lt'
        ts_range[end_timestamp_op] = end
    return ts_range


def make_query_from_filter(sample_filter, require_meter=True):
    """Return a query dictionary based on the settings in the filter.

    :param sample_filter: SampleFilter instance
    :param require_meter: If true and the filter does not have a meter,
                          raise an error.
    """
    q = {}

    """If the length of query counter is more than 1
    then return like '$in': ['counter1', 'counter1']
    """
    if sample_filter.get('meter'):
        counter_names = sample_filter.get('meter').split(',')
        if len(counter_names) > 1:
            q_counter_name = {}
            q_counter_name['$in'] = counter_names
            q['counter_name'] = sample_filter.get('meter')
        else:
            q['counter_name'] = sample_filter.get('meter')
    elif require_meter:
        raise RuntimeError('Missing required meter specifier')

    ts_range = make_timestamp_range(sample_filter.get('start_timestamp'),
                                    sample_filter.get('end_timestamp'),
                                    sample_filter.get('start_timestamp_op'),
                                    sample_filter.get('end_timestamp_op'))

    if ts_range:
        q['timestamp'] = ts_range

    if sample_filter.get('resource'):
        q['resource_id'] = sample_filter.get('resource')



#    Copyright  2017 EasyStack, Inc
#    Author : Branty(jun.wang@easystack.cn)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

import datetime
import json
from oslo_utils import timeutils
import re


AVALIABLE_STATUS = [
    'SHUTOFF',
    'ACTIVE'
    ]


class Singleton(object):

    def __init__(self, clz):
        """Return the last handler which cloud be required more than once."""
        self.clz = clz
        self.instance = None

    def __call__(self, *args, **kwargs):
        if self.instance is None:
            self.instance = self.clz(*args, **kwargs)
        return self.instance


def isUseable_instance(status):
    """
    :param status: nova instance status
    """
    return status in AVALIABLE_STATUS


def is_active(instance):
    """
    :param instance: a nova instance,normally is a dict
    """
    if isinstance(instance, dict):
        return instance['server']['status'] == 'ACTIVE' \
               if "server" in instance.keys() \
               else instance['status'] == "ACTIVE"
    else:
        return False


def endswith_words(source):
    """
    Determine whether a string ends with the pattern of vd[a-z]
    example:
        string = 'aa0d0c92-31a8-44a2-vsfd' =>>> return False
        string = 'aa0d0a-4733-944bfe7-vda' =>>> return True
    :param source: str

    """
    match = False
    if isinstance(source, str):
        match = re.search(".*-vd[a-z]$", source)
    elif isinstance(source, unicode):
        match = re.search(".*-vd[a-z]$", str(source))
    else:
        return False
    return match


def date2str(date):
    return date.strftime("%Y-%m-%d %H:%M:%S")


def str2date(string):
    return datetime.datetime.strptime(string, "%Y-%m-%d %H:%M:%S")


def ms2str(date):
    return date.strftime("%Y-%m-%d %H:%M:%S.%f")


def utcnow():
    """Returns a datetime for the current utc time."""
    return timeutils.utcnow()


def mapping_json_to_dict(mapping_file):
    """
    Parsing the JSON file according to configuration
    The configuration is supported as follows:
    {
     "period_colls":[60,300,3600]
     "60":{
           "meter_type":["cpu_util",
                         "disk.read.bytes.rate",
                         "..."]
            "mult_topology":[1,5,15,120,1440]
            "point_topology":[100,300,100,100,200]
          },
     "300":{
          },
     "3600":{
            "meter_type":["instance",
                          "volume",
                          "account"
                          ...
                        ]
            "mult_topology":[1,6,24]
            "point_topology":[100,200,200]
          }
    }
    """

    def _parse_json_file(j_file):
        with open(j_file) as f:
            return json.load(f)
    try:
        map_dict = _parse_json_file(mapping_file)
        if 'period_colls' in map_dict and \
                isinstance(map_dict['period_colls'], list):
            for i in map_dict['period_colls']:
                if str(i) not in map_dict.keys():
                    raise Exception("Parsing mapping.json error, "
                                    "because of not key %d in "
                                    "the mapping file" % i)
        else:
            raise Exception("Parsing mapping.json error,"
                            "Maybe not key period_colls "
                            "in the mapping file or "
                            "the value of period_colls is not list")
        return map_dict
    except ValueError:
        raise
    except Exception:
        raise


def get_metric_BASE_T(map_dict, metric=None):
    """
    :param map_dict: Parsed mapping.json as a dict

    """
    if not isinstance(map_dict, dict):
        raise
    if metric is None:
        return
    for period in map_dict['period_colls']:
        metrics = map_dict.get(str(period))
        if not metrics:
            continue
        elif metric in metrics['meter_type']:
            return int(period)
    return



#    Copyright  2017 EasyStack, Inc
#    Authors: Hanxi Liu<apolloliuhx@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pika
import time

from eszcp.common import conf
from eszcp.common import log


LOG = log.logger(__name__)
cfg = conf.Conf()

hosts = cfg.read_option('os_rabbitmq', 'rabbit_hosts')
user = cfg.read_option('os_rabbitmq', 'rabbit_user')
passwd = cfg.read_option('os_rabbitmq', 'rabbit_pass')
port = cfg.read_option('os_rabbitmq', 'rabbit_port')
vh = cfg.read_option('os_rabbitmq', 'rabbit_virtual_host')
max_retries = int(cfg.read_option('os_rabbitmq', 'max_retries', -1))
retry_interval = int(cfg.read_option('os_rabbitmq', 'retry_interval', 5))


def connection():
    connect = None
    connection_state = False
    attemps = 0
    MAX_RETRIES = max_retries * len(hosts.split(','))
    while True:
        if connection_state:
            break
        try:
            for host in hosts.split(','):
                LOG.info("Connecting to Rabbitmq server %s..." % host)
                connect = pika.BlockingConnection(pika.ConnectionParameters(
                    host=host,
                    port=int(port),
                    virtual_host=vh,
                    credentials=pika.PlainCredentials(user,
                                                      passwd)))
        except Exception as e:
            if max_retries < 0:
                LOG.error('Unable to connect to the Rabbitmq cluster: '
                          '%(msg)s.Trying again in %(retry_interval)d '
                          'seconds,Continuing to retry to connect '
                          % {'msg': e,
                             'retry_interval': retry_interval})
                time.sleep(retry_interval)
            elif max_retries > 0 and attemps <= MAX_RETRIES:
                LOG.error('Unable to connect to the Rabbitmq cluster: '
                          '%(msg)s.Trying again in %(retry_interval)d '
                          'seconds,max_retries time: %(max_retries)d,'
                          'retry times left:%(left)d'
                          % {'msg': e,
                             'retry_interval': retry_interval,
                             'max_retries': MAX_RETRIES,
                             'left': (MAX_RETRIES - attemps)})
                attemps += 1
                time.sleep(retry_interval)
            else:
                LOG.error('Unable to connect to the Rabbitmq cluster: '
                          '%(msg)s.' % {'msg': e})
                raise
        else:
            connection_state = True
    return connect


class MQConnection(object):
    """RabbitMQ connection class
    """
    def __init__(self):
        self.connection = connection()

    def __call__(self):
        self.connection = connection()



#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#             Hanxi Liu<apolloliuhx@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Zabbix Handler

Provides a class responsible for the communication with Zabbix,

including access to several API methods
"""

import functools
import json
import socket
import struct
import urllib2

from eszcp.common import log
from eszcp import exceptions
from eszcp import utils


LOG = log.logger(__name__)


def logged(func):

    @functools.wraps(func)
    def with_logging(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception, ex:
            LOG.error(ex)
            raise
    return with_logging


class Base_Handler(object):
    def __init__(self, zabbix_host, zabbix_port=10051):
        self.zabbix_host = zabbix_host
        self.zabbix_port = zabbix_port

    def set_proxy_header(self, data):
        """Method used to simplify constructing the protocol to
        communicate with Zabbix

        :param data: refers to the json message
        :rtype : returns the message ready to send to Zabbix server
        with the right header
        """
        # data_length = len(data)
        # data_header = struct.pack('i', data_length) + '\0\0\0\0'
        # HEADER = '''ZBXD\1%s%s'''
        # data_to_send = HEADER % (data_header, data)
        payload = json.dumps(data)
        return payload

    def connect_zabbix(self, payload):
        """
        Method used to send information to Zabbix
        :param payload: refers to the json message prepared to send to Zabbix
        :rtype : returns the response received by the Zabbix API
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.zabbix_host, int(self.zabbix_port)))
        s.send(payload)
        # read its response, the first five bytes are the header again
        response_header = s.recv(5, socket.MSG_WAITALL)
        if not response_header == 'ZBXD\1':
            raise ValueError('Got invalid response')

        # read the data header to get the length of the response
        response_data_header = s.recv(8, socket.MSG_WAITALL)
        response_data_header = response_data_header[:4]
        response_len = struct.unpack('i', response_data_header)[0]

        # read the whole rest of the response now that we know the length
        response_raw = s.recv(response_len, socket.MSG_WAITALL)
        s.close()
        LOG.info(response_raw)
        response = json.loads(response_raw)

        return response

    def send_data_zabbix(self, counter_volume, resource_id,
                         item_key, proxy_name):
        """Method used to prepare the body with data from Ceilometer and send
        it to Zabbix using connect_zabbix method

        :param counter_volume: the actual measurement
        :param resource_id:  refers to the resource ID
        :param item_key:    refers to the item key
        """
        tmp = json.dumps(counter_volume)
        data = {"request": "history data", "host": proxy_name,
                "data": [{"host": resource_id,
                          "key": item_key,
                          "value": tmp}]}

        payload = self.set_proxy_header(data)
        self.connect_zabbix(payload)


class ZabbixHandler(Base_Handler):
    def __init__(self, zabbix_admin_user, zabbix_admin_pass,
                 zabbix_host, zabbix_port,
                 template_name, ks_client, nv_client):
        """
        :param zabbix_user: zabbix admin user
        :param zabbix_pass: zabbix admin password
        :param zabbix_pass: zabbix host ip
        :param template_name:zabbix templete which binds nova instance
        :parms ks_client: Openstack identity service, keystone client
        :parms nv_client: Openstack compute service,nova client
        :parms zabbix_hdl:zabbix handler
        """
        super(ZabbixHandler, self).__init__(zabbix_host, zabbix_port)
        self.zabbix_admin_user = zabbix_admin_user
        self.zabbix_admin_pass = zabbix_admin_pass
        self.template_name = template_name
        self.ks_client = ks_client
        self.nv_client = nv_client
        self.group_list = []

    def first_run(self):

        self.api_auth = self.get_zabbix_auth()
        self.proxies = self.check_proxies()
        self.template_id = self.get_template_id()
        tenants = self.ks_client.get_projects()
        self.group_list = self.host_group_list(tenants)
        self.check_host_groups()
        self.check_instances()

    def get_zabbix_auth(self):
        """
        Method used to request a session ID form Zabbix API by sending
        Admin credentials (user, password)

        :return: returns an Id to use with zabbix api calls
        """
        payload = {"jsonrpc": "2.0",
                   "method": "user.login",
                   "params": {"user": self.zabbix_admin_user,
                              "password": self.zabbix_admin_pass},
                   "id": 2}
        response = self.contact_zabbix_server(payload)
        if 'error' in response:
            LOG.error('Incorrect user or password,please check it again')
            raise
        zabbix_auth = response['result']
        return zabbix_auth

    def check_proxies(self):
        """
        Method used to check if those proxies exist.
        Map keystone domain to zabbix proxy.

        :return: a control value and those proxies if exists
        """
        cache_proxy = {}
        payload = {
            "jsonrpc": "2.0",
            "method": "proxy.get",
            "params": {
                "output": "extend"
            },
            "auth": self.api_auth,
            "id": 1
        }
        # return all proxies in zabbix
        response = self.contact_zabbix_server(payload)
        proxies = [proxy['host'] for proxy in response['result']]
        for proxy in response['result']:
            cache_proxy[proxy['host']] = proxy['proxyid']
        # return all domains in keystone
        domains = [domain for domain in self.ks_client.get_domains()]
        # each item in current_proxies:[proxy_id,domain_id]
        current_proxies = []
        for domain in domains:
            if domain.name not in proxies:
                '''
                Check if proxy exists, if not create one
                '''
                # host: domain ID truncated by the first eight
                if domain.id[:8] in proxies:
                    current_proxies.append([cache_proxy[domain.id[:8]],
                                            domain.id])
                    continue
                LOG.info("%s is not in zabbix proxies, starting to create a "
                         "new proxy mapping to keystone domain" % domain.name)
                payload = {
                       "jsonrpc": "2.0",
                       "method": "proxy.create",
                       "params": {
                           "host": domain.name,
                           "status": "5"
                       },
                       "auth": self.api_auth,
                       "id": 1
                       }
                response = self.contact_zabbix_server(payload)
                # NOTE: when '@','#' or other No ASSIC char in name string,
                # raise 'Invalid Params' error.
                # Then Replace zabbix proxy name with domain ID
                # host: domain ID truncated by the first eight
                if 'error' in response:
                    payload['params']['host'] = domain.id[:8]
                    response = self.contact_zabbix_server(payload)
                proxy_id = response['result']['proxyids'][0]
                current_proxies.append([proxy_id, domain.id])
                LOG.info("%s is not in zabbix proxies, success to create a "
                         "new proxy mapping to keystone domain" % domain.name)
            else:
                current_proxies.append([[cache_proxy[domain.name]], domain.id])
        return current_proxies

    @logged
    def check_host_groups(self):
        """
        This method checks if some host group exists

        """
        payload = {
                "jsonrpc": "2.0",
                "method": "hostgroup.get",
                "params": {
                    "output": "extend"
                },
                "auth": self.api_auth,
                "id": 1
        }
        # Get all host groups
        response = self.contact_zabbix_server(payload)
        zabbix_hostgroups = [gp['name'] for gp in response['result']]
        for item in self.group_list:
            if item[0] not in zabbix_hostgroups:
                # Prevent naming collision
                if str(item[0] + item[1][:6]) in zabbix_hostgroups:
                    continue
                payload = {"jsonrpc": "2.0",
                           "method": "hostgroup.create",
                           "params": {"name": item[0]},
                           "auth": self.api_auth,
                           "id": 2}
                response = self.contact_zabbix_server(payload)
                if 'error' in response:
                    # NOTE:host name collision ,rename host_name =
                    # <tenant ID truncated by the first six> + <tenant_nam>
                    payload['params']['name'] = item[0] + item[1][:6]
                    response = self.contact_zabbix_server(payload)
                if response.get('result'):
                    LOG.info("Success to create a new hostgroup: %s, "
                             "zabbix hostgroup mapping to keystone project."
                             % payload['params']['name'])
                else:
                    LOG.info("Failed to create a new hostgroup: %s, "
                             "zabbix hostgroup mapping to keystone project."
                             % payload['params']['name'])
            else:
                LOG.info("Already exists a new hostgroup: %s, "
                         "zabbix hostgroup mapping to keystone project."
                         % item[0])

    @logged
    def check_instances(self):
        """
        Method used to verify existence of an instance / host

        """
        servers = [i.to_dict() for i in self.nv_client.instance_get_all()]
        for item in servers:
            if utils.isUseable_instance(item['status']):
                payload = {
                    "jsonrpc": "2.0",
                    "method": "host.get",
                    "params": {
                        "output": "extend",
                        "filter": {"host": item['id']}
                        },
                    "auth": self.api_auth,
                    "id": 1
                }
                response = self.contact_zabbix_server(payload)
                # the Nova instance has not been create in zabbix,
                # so the result in the return body in []
                if not response.get('result', []):
                    for row in self.group_list:
                        if row[1] == item['tenant_id']:
                            instance_name = item['name']
                            instance_id = item['id']
                            domain_id = row[2]
                            tenant_name = row[0]
                            self.create_host(
                                         instance_name,
                                         instance_id,
                                         tenant_name,
                                         item['tenant_id'],
                                         domain_id,
                                         )
                            LOG.info(
                                "Success to create a new host: %s, "
                                "The instances is to a keystone "
                                "project(%s),and now create a new zabbix host "
                                "added to the zabbix hostgroup. "
                                % (instance_name,
                                   tenant_name
                                   )
                            )
                else:
                    LOG.info("Zabbix host(%s) has already exists."
                             % item['id'])
            else:
                msg = "Drop to check or create instance ," \
                      "the status of %(instance_name)s(%(instance_id)s) " \
                      "is in %(status)s." \
                      % {"instance_name": item['name'],
                         "instance_id": item['id'],
                         "status": item['status']}
                LOG.warning(msg)

    def create_host(self, instance_name, instance_id,
                    tenant_name, tenant_id, domain_id=None):

        """
        Method used to create a host in Zabbix server

        :param instance_name: refers to the instance name
        :param instance_id:   refers to the instance id
        :param tenant_name:   refers to the tenant name
        :param domain_id:     refers to the domain id
        """
        group_id = self.find_group_id(tenant_name, tenant_id)
        if not domain_id:
            project = self.ks_client.get_project(tenant_id)
            domain_id = project.domain_id
        proxy_id = self.find_proxy_id(domain_id)
        hostname = instance_name
        if not (instance_id in instance_name):
            hostname = instance_name + '_' + instance_id[:8]

        payload = {"jsonrpc": "2.0",
                   "method": "host.create",
                   "params": {
                       "host": instance_id,
                       "name": hostname,
                       "proxy_hostid": proxy_id,
                       "interfaces": [
                           {
                               "type": 1,
                               "main": 1,
                               "useip": 1,
                               "ip": "127.0.0.1",
                               "dns": "",
                               "port": "10050"}
                       ],
                       "groups": [
                           {
                               "groupid": group_id
                           }
                       ],
                       "templates": [
                           {
                               "templateid": self.template_id
                           }
                       ],

                   },
                   "auth": self.api_auth,
                   "id": 1}
        response = self.contact_zabbix_server(payload)
        if response.get('result'):
            LOG.info("Success to create a new host: %s" % instance_name)
        else:
            LOG.warning("Failed to create a new host: %s" % instance_name)

    def find_group_id(self, tenant_name, tenant_id):
        """
        Method used to find the the group id of an host in Zabbix server

        :param tenant_name: refers to the tenant name
        :param tenant_id: refers to the tenant id
        :return: returns the group id that belongs to the host_group or tenant
        """
        group_id = None
        payload = {"jsonrpc": "2.0",
                   "method": "hostgroup.get",
                   "params": {
                       "output": "extend"
                   },
                   "auth": self.api_auth,
                   "id": 2
                   }
        response = self.contact_zabbix_server(payload)
        group_list = response['result']
        for line in group_list:
            if line['name'] == tenant_name:
                group_id = line['groupid']
                break
            elif line['name'] == tenant_name + tenant_id[:6]:
                group_id = line['groupid']
                break
        return group_id

    def find_proxy_id(self, domain_id):
        """
        Method used to find the the proxy id of an host in Zabbix server
        :param domain_name: refers to the domain name
        :param domain_id: refers to the domain id
        :return: returns the proxy id that belongs to the zabbix proxy
        """
        # When domain_id not found, raise exception
        # TO DO
        domain = self.ks_client.show_domain(domain_id)
        payload = {"jsonrpc": "2.0",
                   "method": "proxy.get",
                   "params": {
                       "output": "extend",
                       "filter": {
                            "host": domain.name
                           }
                   },
                   "auth": self.api_auth,
                   "id": 2
                   }
        response = self.contact_zabbix_server(payload)
        proxy = response['result']
        if proxy and len(proxy) > 0:
            return proxy[0]['proxyid']
        else:
            payload['params']['filter']['host'] = domain.id[:8]
            response = self.contact_zabbix_server(payload)
            proxy = response['result']
            if proxy and len(proxy) > 0:
                return proxy[0]['proxyid']
        # proxy not found
        raise

    def get_template_id(self):
        """
        Method used to check if the template already exists.
        If not, creates one

        :return: returns the template ID
        """
        global template_id
        template_id = None
        payload = {
            "jsonrpc": "2.0",
            "method": "template.get",
            "params": {
                "output": "extend",
                "filter": {
                    "host": [
                        self.template_name
                    ]
                }
            },
            "auth": self.api_auth,
            "id": 1
        }
        response = self.contact_zabbix_server(payload)

        if len(response['result']) > 0:
            for item in response['result']:
                template_id = item['templateid']
        else:
            message = ("Can't find default template in zabbix. "
                       "Please import the default template!")
            raise exceptions.TemplateNotFound(message)
        return template_id

    def find_host_id(self, host):
        """
        Method used to find a host Id in Zabbix server

        :param host: ref to nova instance_id
        :return: returns the host id
        """
        host_id = None
        payload = {"jsonrpc": "2.0",
                   "method": "host.get",
                   "params": {
                       "output": "extend",
                       "filter": {"host": host}
                   },
                   "auth": self.api_auth,
                   "id": 2
                   }
        response = self.contact_zabbix_server(payload)
        if 'error' in response:
            LOG.error("Falied to find host(%s):%s,maybe zabbix server "
                      "is down" % (host, response['error']['data']))
        elif len(response.get('result')):
            host_id = response['result'][0]['hostid']
        else:
            LOG.waring("Falied to find host: %s" % host)
        return host_id

    def delete_host(self, host_id):
        """
        Method used to delete a Host in Zabbix Server

        :param host_id: refers to the host id to delete
        """
        payload = {"jsonrpc": "2.0",
                   "method": "host.delete",
                   "params": [
                       host_id
                   ],
                   "auth": self.api_auth,
                   "id": 1
                   }
        response = self.contact_zabbix_server(payload)
        if 'error' in response:
            LOG.error("Falied to delete host:%s,maybe zabbix server "
                      "is down" % host_id)
        else:
            LOG.info("Success to delete host:%s" % host_id)

    def host_group_list(self, tenants):
        """
        Method to "fill" an array of hosts

        :param tenants: receive an array of tenants
        :return: parsed list of hosts [[tenant_name1, uuid1],
        [tenant_name2, uuid2], ..., [tenant_nameN, uuidN],]
        """
        host_group_list = []
        for item in tenants:
            if not item.name == 'services':
                host_group_list.append([item.name,
                                        item.id,
                                        item.domain_id
                                        ]
                                       )

        return host_group_list

    def project_delete(self, tenant_id):
        """
        Method used to delete a project

        :param tenant_id: receives a tenant id
        """

        for item in self.group_list:
            if item[1] == tenant_id:
                tenant_name = item[0]
                group_id = self.find_group_id(tenant_name, tenant_id)
                self.delete_host_group(group_id)
                self.group_list.remove(item)

    def delete_host_group(self, group_id):
        """
        This method deletes a host group
        :param group_id: receives the group id
        """
        payload = {"jsonrpc": "2.0",
                   "method": "hostgroup.delete",
                   "params": [group_id],
                   "auth": self.api_auth,
                   "id": 1
                   }
        self.contact_zabbix_server(payload)

    def create_host_group(self, tenant_name):
        """
        This method is used to create host_groups. Every tenant is a host group

        :param tenant_name: receives keystone tenant name
        """
        payload = {"jsonrpc": "2.0",
                   "method": "hostgroup.create",
                   "params": {"name": tenant_name},
                   "auth": self.api_auth,
                   "id": 2}
        self.contact_zabbix_server(payload)

    def create_proxy(self, domain_name, domain_id):
        """
        This method is used to create zabbix proxy. Every domain
        is a zabbix proxy

        :param domain_name: refs to keystone domain name
        :param domain_id:   refs to keystone domain id
        """
        payload = {
                       "jsonrpc": "2.0",
                       "method": "proxy.create",
                       "params": {
                           "host": domain_name,
                           "status": "5"
                       },
                       "auth": self.api_auth,
                       "id": 1
        }
        proxy_id = None
        response = self.contact_zabbix_server(payload)
        # NOTE: when '@','#' or other No ASSIC char in name string,
        # raise 'Invalid Params' error.
        # Then Replace zabbix proxy name with domain ID
        # host: domain ID truncated by the first eight
        if 'error' in response:
            payload['params']['host'] = domain_id[:8]
            response = self.contact_zabbix_server(payload)
            if 'error' not in response:
                proxy_id = response['result']['proxyids'][0]
                LOG.info("Success to create a new proxy: %s" % domain_id[:8])
                self.proxies.append([proxy_id, domain_id])
            else:
                LOG.warning("Failed to create a new proxy: %s,"
                            "error message: %s" % (domain_id[:8],
                                                   response['error']))
        elif response.get('result'):
            proxy_id = response['result']['proxyids'][0]
            LOG.info("Success to create a new proxy: %s" % domain_name)
            self.proxies.append([proxy_id, domain_id])

    def delete_proxy(self, domain_id):
        for item in self.proxies:
            if item[1] == domain_id:
                payload = {
                    "jsonrpc": "2.0",
                    "method": "proxy.delete",
                    "params": [item[0]],
                    "auth": self.api_auth,
                    "id": 1
                }
                response = self.contact_zabbix_server(payload)
                if 'result' in response:
                    LOG.info("Success to delete a new proxy: %s"
                             % response['result']['proxyids'][0])
                    self.proxies.remove(item)
                else:
                    LOG.warning("Falied to delete a new proxy, error "
                                "message:%s" % response['error'])

    def get_items_by_template(self):
        """This method find all items in specified template
        :return : a list stored all metric items
        """
        template_id = self.get_template_id()
        payload = {"jsonrpc": "2.0",
                   "method": "item.get",
                   "params": {"output": "extend",
                              "templateids": [template_id]},
                   "auth": self.api_auth,
                   "id": 1
                   }
        response = self.contact_zabbix_server(payload)
        if 'error' in response:
            return []
        else:
            return [item['name'] for item in response['result']]

    def get_hosts_by_proxy(self, proxy_id=None):
        """This method find all hosts in specified proxy
        :return : a dict stored all hosts
        """
        hosts = {}
        # proxy_name is not None,find all hosts in the specified proxy
        if proxy_id:
            payload = {"jsonrpc": "2.0",
                       "method": "host.get",
                       "params": {"output": "extend",
                                  "proxyids": [proxy_id]},
                       "auth": self.api_auth,
                       "id": 1
                       }
            response = self.contact_zabbix_server(payload)
            if 'error' not in response:
                hosts = {proxy_id: response['result']}
        else:
            for i in self.proxies:
                payload = {
                    "jsonrpc": "2.0",
                    "method": "host.get",
                    "params": {"output": "extend",
                               "proxyids": [i[0]]},
                    "auth": self.api_auth,
                    "id": 1
                   }
                response = self.contact_zabbix_server(payload)
                hosts[i[0]] = [(host['host'], host['hostid'], host['name'])
                               for host in response['result']]
        return hosts

    def get_hosts(self, filter_no_proxy=False):
        """Method used do query Zabbix API in order to fill an Array of hosts

        :params filter_no_proxy: get all hosts whose proxy_hostid is not '0'
        :return: returns a dict and a list ,record zabbix host information.
        hosts_cache, the date structure is the following:
         {"instance_id":["host_id","host_name","proxy_id"]
         }
        hosts: ['instance_id1','instance)id2',...]
        """
        hosts = []
        hosts_cache = {}
        payload = {
                    "jsonrpc": "2.0",
                    "method": "host.get",
                    "params": {"output": "extend"},
                    "auth": self.api_auth,
                    "id": 1
        }
        response = self.contact_zabbix_server(payload)
        if 'error' in response:
            LOG.warning("Falied to get all zabbix hosts, error "
                        "message:%s" % response['error'])
            return []
        # Get hosts monitoring the zabbix proxy
        if filter_no_proxy:
            for host in response['result']:
                if int(host['proxy_hostid']):
                    hosts.append(host['host'])
                    hosts_cache[host['host']] = [host['hostid'],
                                                 host['name'],
                                                 host['proxy_hostid']]
        else:
            for host in response['result']:
                hosts.append(host['host'])
                hosts_cache[host['host']] = [host['hostid'],
                                             host['name'],
                                             host['proxy_hostid']]
        return hosts, hosts_cache

    def get_by_proxyid(self, proxy_id):
        """
        :param proxy_id: refs to proxy id
        """
        payload = {
                    "jsonrpc": "2.0",
                    "method": "proxy.get",
                    "params": {"output": "extend",
                               "proxyids": proxy_id},
                    "auth": self.api_auth,
                    "id": 1
        }
        response = self.contact_zabbix_server(payload)
        if 'error' in response:
            LOG.error("Falied to find proxy : %s" % proxy_id)
        elif len(response['result']) == 0:
            LOG.warning("Can't find proxy : %s" % proxy_id)
        else:
            return response['result'][0]

    def contact_zabbix_server(self, payload):
        """
        Method used to contact the Zabbix server.

        :param payload: refers to the json message to send to Zabbix
        :return: returns the response from the Zabbix API
        """
        data = json.dumps(payload)
        req = urllib2.Request('http://' + self.zabbix_host +
                              '/zabbix/api_jsonrpc.php',
                              data,
                              {'Content-Type': 'application/json'})
        f = urllib2.urlopen(req)
        response = json.loads(f.read())
        f.close()
        return response


#    Copyright  2017 EasyStack, Inc
#    Authors:  Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

INSTANCE_METRICS = [
    'cpu_util',
    'cpu.delta',
    'memory.usage',
    'disk.read.bytes.rate',
    'disk.read.requests.rate',
    'disk.write.bytes.rate',
    'disk.write.requests.rate'
    ]

NETWORK_METRICS = [
    'network.incoming.bytes.rate',
    'network.incoming.packets.rate',
    'network.outgoing.bytes.rate',
    'network.outgoing.packets.rate'
    ]

"""
 Cache instance metrics, the date structure is the following:
 {"instance_id":{
     "instance_id": INSTANCE_METRICS,
     "instance-xxx-{instance_id}-{tap_id}": NETWORK_METRICS,
     "instance-{disk_id}": DISK_METRICS,
     ...
    },
  ...
 }
"""



#    Copyright  2017 EasyStack, Inc
#    Authors: Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import six


class Model(object):
    """Base class for storage API models."""

    def __init__(self, **kwds):
        self.fields = list(kwds)
        for k, v in six.iteritems(kwds):
            setattr(self, k, v)

    def as_dict(self):
        d = {}
        for f in self.fields:
            v = getattr(self, f)
            if isinstance(v, Model):
                v = v.as_dict()
            elif isinstance(v, list) and v and isinstance(v[0], Model):
                v = [sub.as_dict() for sub in v]
            d[f] = v
        return d


class Resource(Model):
    """Something for which sample data has been collected."""

    def __init__(self, resource_id, project_id,
                 first_sample_timestamp,
                 last_sample_timestamp,
                 source, user_id, metadata,
                 resource_name):
        """Create a new resource.

        :param resource_id: UUID of the resource
        :param project_id:  UUID of project owning the resource
        :param first_sample_timestamp: first sample timestamp captured
        :param last_sample_timestamp: last sample timestamp captured
        :param source:      the identifier for the user/project id definition
        :param user_id:     UUID of user owning the resource
        :param metadata:    most current metadata for the resource (a dict)
        :param resource_name:   resource display name
        """
        super(Resource, self).__init__(
            resource_id=resource_id,
            first_sample_timestamp=first_sample_timestamp,
            last_sample_timestamp=last_sample_timestamp,
            project_id=project_id,
            source=source,
            user_id=user_id,
            metadata=metadata,
            resource_name=resource_name)


class Statistics(Model):
    """Computed statistics based on a set of sample data."""
    def __init__(self, unit,
                 period, period_start, period_end,
                 duration, duration_start, duration_end,
                 groupby, **data):
        """Create a new statistics object.

        :param unit: The unit type of the data set
        :param period: The length of the time range covered by these stats
        :param period_start: The timestamp for the start of the period
        :param period_end: The timestamp for the end of the period
        :param duration: The total time for the matching samples
        :param duration_start: The earliest time for the matching samples
        :param duration_end: The latest time for the matching samples
        :param groupby: The fields used to group the samples.
        :param data: some or all of the following aggregates
           min: The smallest volume found
           max: The largest volume found
           avg: The average of all volumes found
           sum: The total of all volumes found
           count: The number of samples found
           aggregate: name-value pairs for selectable aggregates
        """
        super(Statistics, self).__init__(
            unit=unit,
            period=period,
            period_start=period_start,
            period_end=period_end,
            duration=duration,
            duration_start=duration_start,
            duration_end=duration_end,
            groupby=groupby,
            **data)



#    Copyright  2017 EasyStack, Inc
#    Authors: Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import datetime
import os
import pymongo
import six
import time

from eszcp import exceptions
from eszcp.common import conf
from eszcp.common import log
from eszcp.common.db import models
from eszcp.common.db import pymongo_utils
from eszcp import utils

LOG = log.logger(__name__)
CONF = conf.Conf()

MAX_RETRIES = CONF.read_option('mongodb', 'max_retries', 3)
RETRY_INTERVAL = CONF.read_option('mongodb', 'retry_interval', 2)
MONGO_URL = CONF.read_option('mongodb', 'connection')
MAPPING_FILE = CONF.read_option('mongodb', 'mapping_file',
                                '/etc/ceilometer/mapping.json')
CACHE_MAPPING_FILE = {}


def parse_metric_json():
    global CACHE_MAPPING_FILE
    mapping_file = MAPPING_FILE
    if not os.path.isabs(mapping_file):
        CACHE_MAPPING_FILE = '/etc/ceilometer/%s' % mapping_file
    elif os.path.exists(mapping_file):
        CACHE_MAPPING_FILE = utils.mapping_json_to_dict(mapping_file)
    if not CACHE_MAPPING_FILE:
        LOG.error("Can't find Metric mapping.json file, "
                  "Make sure the mapping_file exist "
                  "under section [collector] "
                  "which must be configured properly.")
        raise exceptions.MappingFileNotFound
    return CACHE_MAPPING_FILE


def safe_mongo_call(call):
    def closure(*args, **kwargs):
        max_retries = MAX_RETRIES
        retry_interval = RETRY_INTERVAL
        attempts = 0
        while True:
            try:
                return call(*args, **kwargs)
            except pymongo.errors.AutoReconnect as err:
                if 0 <= max_retries <= attempts:
                    LOG.error('Unable to reconnect to the primary mongodb '
                              'after %(retries)d retries. Giving up.' %
                              {'retries': max_retries})
                    raise
                LOG.warning('Unable to reconnect to the primary mongodb: '
                            '%(errmsg)s. Trying again in %(retry_interval)d '
                            'seconds.' %
                            {'errmsg': err, 'retry_interval': retry_interval})
                attempts += 1
                time.sleep(retry_interval)
    return closure


@utils.Singleton
class Connection(object):
    """MongoDB connection.
    """
    SORT_OPERATION_MAPPING = {'desc': pymongo.DESCENDING,
                              'asc': pymongo.ASCENDING}
    _GENESIS = datetime.datetime(year=datetime.MINYEAR, month=1, day=1)
    _APOCALYPSE = datetime.datetime(year=datetime.MAXYEAR, month=12, day=31,
                                    hour=23, minute=59, second=59)
    SAMPLE_T = 60

    def __init__(self):
        url = MONGO_URL
        max_retries = MAX_RETRIES
        retry_interval = RETRY_INTERVAL
        attempts = 0
        while True:
            try:
                self.client = pymongo.MongoClient(url)
                LOG.debug('mongo client: %s' % self.client)
                parse_metric_json()
            except pymongo.errors.ConnectionFailure as e:
                if max_retries >= 0 and attempts >= max_retries:
                    LOG.error('Unable to connect to the database after '
                              '%(retries)d retries. Giving up.' %
                              {'retries': max_retries})
                    raise
                LOG.warning('Unable to connect to the database server: '
                            '%(errmsg)s. Trying again in %(retry_interval)d '
                            'seconds.' %
                            {'errmsg': e, 'retry_interval': retry_interval})
                attempts += 1
                time.sleep(retry_interval)
            except Exception as e:
                LOG.warning('Unable to connect to the database server: '
                            '%(errmsg)s.' % {'errmsg': e})
                raise
            else:
                connection_options = pymongo.uri_parser.parse_uri(url)
                self.db = getattr(self.client, connection_options['database'])
                self.db.authenticate(connection_options['username'],
                                     connection_options['password'])
                break

    def get_resources(self, start_timestamp=None, start_timestamp_op=None,
                      end_timestamp=None, end_timestamp_op=None,
                      metaquery=None, resource=None, limit=None):
        """Return an iterable of models.Resource instances
        :param start_timestamp: Optional modified timestamp start range.
        :param start_timestamp_op: Optional start time operator, like gt, ge.
        :param end_timestamp: Optional modified timestamp end range.
        :param end_timestamp_op: Optional end time operator, like lt, le.
        :param metaquery: Optional dict with metadata to match on.
        :param resource: Optional resource filter.
        :param limit: Maximum number of results to return.
        """
        metaquery = pymongo_utils.improve_keys(metaquery, metaquery=True) or {}
        if start_timestamp or end_timestamp:
            # TO DO
            # ceilometet.storage.impl_mongo.ConnectionOrig._get_time_constrained_resources
            return []
        else:
            query = {}
            sort_instructions = []
            if resource is not None:
                query['_id'] = resource
            query.update(dict((k, v)
                         for (k, v) in six.iteritems(metaquery)))
            sort_key = 'last_sample_timestamp'
            sort_instructions.append((sort_key,
                                     self.SORT_OPERATION_MAPPING['desc']))
            if limit:
                results = self.db.resource.find(query, sort=sort_instructions,
                                                limit=limit)
            else:
                results = self.db.resource.find(query, sort=sort_instructions)

            return [models.Resource(
                        resource_id=r['_id'],
                        user_id=r['user_id'],
                        project_id=r['project_id'],
                        first_sample_timestamp=r.get('first_sample_timestamp',
                                                     self._GENESIS),
                        last_sample_timestamp=r.get('last_sample_timestamp',
                                                    self._APOCALYPSE),
                        source=r['source'],
                        metadata=pymongo_utils.unquote_keys(r['metadata']),
                        resource_name=pymongo_utils.unquote_keys(
                                   r['resource_name'])) for r in results]

    def get_meter_statistics(self, sample_filter, period=None, groupby=None,
                             aggregate=None, limit=None):
        """Return an iterable of models.Statistics instance containing meter
        statistics described by the query parameters.

        The filter must have a meter value set.

        """
        if groupby or aggregate:
            # TO DO
            # ceilometet.storage.impl_mongo.Connection.get_meter_statistics
            return []
        period = []
        aggregate = []
        if (groupby and
                set(groupby) - set(['user_id', 'project_id',
                                    'resource_id', 'source'])):
            raise NotImplementedError("Unable to group by these fields")
        q = pymongo_utils.make_query_from_filter(sample_filter)
        if period:
            T = period
        else:
            # Set the smallest base_period as default sample period
            # T = self.SAMPLE_T
            T = utils.get_metric_BASE_T(CACHE_MAPPING_FILE,
                                        sample_filter.get('meter')) \
                                        or self.SAMPLE_T
        coll = 'statistics%s' % T
        LOG.debug("get_statistics2 q = %s" % q)
        if limit:
            results = self.db[coll].find(q,
                                         sort=[('period_start', -1)],
                                         limit=limit)
        else:
            results = self.db[coll].find(q, sort=[('period_start', 1)])

        stats = [self._stats_result_to_model(r, groupby, aggregate)
                 for r in results]
        return stats

    def _stats_result_aggregates(self, result, aggregate):
        stats_args = {}
        for attr in ['count', 'min', 'max', 'sum', 'avg']:
            if attr in result:
                stats_args[attr] = result[attr]

        if aggregate:
            stats_args['aggregate'] = {}
            for a in aggregate:
                ak = '%s%s' % (a.func, '/%s' % a.param if a.param else '')
                if ak in result:
                    stats_args['aggregate'][ak] = result[ak]
                elif 'aggregate' in result:
                    stats_args['aggregate'][ak] = result['aggregate'].get(ak)
        return stats_args

    def _stats_result_to_model(self, result, groupby, aggregate,
                               period=None, first_timestamp=None):

        stats_args = self._stats_result_aggregates(result, aggregate)
        stats_args['unit'] = result['unit']
        stats_args['duration'] = result['T'] if 'T' in result \
            else result['duration']
        stats_args['duration_start'] = result['period_start']
        stats_args['duration_end'] = result['period_end']
        stats_args['period'] = result['T'] if 'T' in result \
            else result['period']
        stats_args['period_start'] = result['period_start']
        stats_args['period_end'] = result['period_end']
        stats_args['groupby'] = (dict(
            (g, result['groupby'][g]) for g in groupby) if groupby else None)
        return models.Statistics(**stats_args)



#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#             Hanxi Liu<apolloliuhx@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Class for Handling KeystoneEvents in OpenStack's RabbitMQ

Uses the pika library for handling the AMQP protocol,

implementing the necessary callbacks for Keystone events
"""

import json

from eszcp.common import log


LOG = log.logger(__name__)


class KeystoneEvents:

    def __init__(self, zabbix_handler, connection, ks_client):
        """
        :param zabbix_handler: zabbix api handler
        :param connection: rabbitmq connection instance
        :param ks_client: keystone client
        """
        self.zabbix_handler = zabbix_handler
        self.ks_client = ks_client
        self.mqconnc = connection

    def keystone_amq(self):
        """
        Method used to listen to keystone events
        """
        try:
            LOG.info("Start consuming keystone messages from rabbitmq ...")
            channel = self.mqconnc.connection.channel()
            channel.exchange_declare(exchange='keystone', type='topic')
            channel.queue_declare(queue="zcp-keystone", exclusive=True)
            channel.queue_bind(exchange='keystone',
                               queue="zcp-keystone",
                               routing_key='notifications.#')
            channel.basic_consume(self.keystone_callback,
                                  queue="zcp-keystone",
                                  no_ack=True)
            channel.start_consuming()
        except Exception as e:
            LOG.error("Fail to consume messages from keystone: %s" % e)
            # Make sure consume messages normally
            self.mqconnc()

    def _handler_events(self, payload):
        if payload['event_type'] == 'identity.project.created':
            tenant_id = payload['payload']['resource_info']
            project = self.ks_client.get_project(tenant_id)
            LOG.info("Creating a hostgroup: %s(%s) in Zabbix Server"
                     % (tenant_id, project.name))
            self.zabbix_handler.create_host_group(project.name)
            self.zabbix_handler.group_list.append([project.name,
                                                   tenant_id,
                                                   project.domain_id])

        elif payload['event_type'] == 'identity.project.deleted':
            tenant_id = payload['payload']['resource_info']
            LOG.info("Deleting a hostgroup: %s in Zabbix Server"
                     % tenant_id)
            self.zabbix_handler.project_delete(tenant_id)
        elif payload['event_type'] == 'identity.domain.created':
            domain_id = payload['payload']['resource_info']
            domain_ref = self.ks_client.show_domain(domain_id)
            domain_name = domain_ref['name'] if isinstance(
                          domain_ref, dict) else domain_ref.name
            LOG.info("Creating a zabbix proxy: %s(%s) in Zabbix Server"
                     % (domain_name, domain_id))
            self.zabbix_handler.create_proxy(domain_name, domain_id)
        elif payload['event_type'] == 'identity.domain.deleted':
            domain_id = payload['payload']['resource_info']
            LOG.info("Deleting a zabbix proxy: %s in Zabbix Server"
                     % domain_id)
            self.zabbix_handler.delete_proxy(domain_id)
        else:
            # TO DO
            # Maybe more event types will be supported
            pass

    def keystone_callback(self, ch, method, properties, body):
        """
        Method used by method keystone_amq() to filter messages
        by type of message.

        :param ch: refers to the head of the protocol
        :param method: refers to the method used in callback
        :param properties: refers to the proprieties of the message
        :param body: refers to the message transmitted
        """
        payload = json.loads(body)
        try:
            self._handler_events(payload)
        except Exception as ex:
            LOG.error(ex)



#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Proxy for integration of resources between OpenStack's Ceilometer and Zabbix

This proxy periodically checks for changes in Ceilometer's resources reporting
them to Zabbix. It is also integrated

OpenStack's Nova and RabbitMQ for reflecting changes in
Projects/Tenants and Instances
"""

import multiprocessing

from eszcp.common import log
from eszcp.common import conf
from eszcp.task.polling.base_handler import HandlerAdapter
from eszcp.task import nova_handler
from eszcp.task import keystone_handler
from eszcp.keystone_client import Client
from eszcp.nova_client import Client as nova_Client
from eszcp import messaging
from eszcp import zabbix_handler

log.initlog()
LOG = log.logger(__name__)
conf_file = conf.Conf()


def init_zcp(processes):
    """
        Method used to initialize the Zabbix-Ceilometer Proxy
    """

    # Creation of the Auth keystone-dedicated authentication class
    # Responsible for managing domain and project related requests
    ks_client = Client(conf_file)
    nv_client = nova_Client(conf_file)
    # Creation of the Zabbix Handler class
    # Responsible for the communication with Zabbix
    zabbix_hdl = zabbix_handler.ZabbixHandler(conf_file.read_option(
                                                    'zabbix_configs',
                                                    'zabbix_admin_user'),
                                              conf_file.read_option(
                                                    'zabbix_configs',
                                                    'zabbix_admin_pass'),
                                              conf_file.read_option(
                                                    'zabbix_configs',
                                                    'zabbix_host'),
                                              conf_file.read_option(
                                                    'zabbix_configs',
                                                    'zabbix_port'),
                                              conf_file.read_option(
                                                    'zcp_configs',
                                                    'template_name'),
                                              ks_client,
                                              nv_client)

    # Creation of the Polling Handler class
    # Responsible for the communication with OpenStack's Ceilometer or Mongo,
    # polling for changes every N seconds
    polling_hdl = HandlerAdapter.get_handler(conf_file, ks_client,
                                             nv_client, zabbix_hdl)

    # First run of the Zabbix handler for retrieving the necessary information
    zabbix_hdl.first_run()

    # Creation of the Nova Handler class
    # Responsible for detecting the creation of new instances in OpenStack,
    # translated then to Hosts in Zabbix
    nova_hdl = nova_handler.NovaEvents(zabbix_hdl, messaging.MQConnection())

    # Creation of the Project Handler class
    # Responsible for detecting the creation of new tenants in OpenStack,
    # translated then to HostGroups in Zabbix
    keystone_hdl = keystone_handler.KeystoneEvents(zabbix_hdl,
                                                   messaging.MQConnection(),
                                                   ks_client)

    # Create and append processes to process list
    LOG.INFO('**************** Keystone listener started ****************')
    p1 = multiprocessing.Process(target=keystone_hdl.keystone_amq)
    p1.daemon = True
    processes.append(p1)

    LOG.INFO('**************** Nova listener started ****************')
    p2 = multiprocessing.Process(target=nova_hdl.nova_amq)
    p2.daemon = True
    processes.append(p2)

    p3 = multiprocessing.Process(target=polling_hdl.interval_run)
    p3.daemon = True
    processes.append(p3)

    # start all the processes
    [ps.start() for ps in processes]


def main():
    processes = []
    LOG.info("-------------- Starting Zabbix Ceilometer Proxy --------------")
    init_zcp(processes)
    # wait for all processes to complete
    [ps.join() for ps in processes]
    LOG.info("************* Runnig Zabbix Ceilometer Proxy  ****************")
    print "ZCP terminated"
    LOG.info('************* Terninate Zabbix Ceilometer Proxy ***************')



#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#             Hanxi Liu<apolloliuhx@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Class for Handling Nova events in OpenStack's RabbitMQ

Uses the pika library for handling the AMQP protocol, implementing the

necessary callbacks for Nova events
"""

import json

from eszcp.common import log


LOG = log.logger(__name__)


class NovaEvents:
    def __init__(self, zabbix_handler, connection):

        """
        :param zabbix_handler: zabbix api handler
        :param connection: rabbitmq connection instance
        """
        self.zabbix_handler = zabbix_handler
        self.mqconnc = connection

    def nova_amq(self):
        """
        Method used to listen to nova events

        """
        try:
            LOG.info("Start consuming nova messages from rabbitmq ...")
            channel = self.mqconnc.connection.channel()
            channel.exchange_declare(exchange='nova', type='topic')
            channel.queue_declare(queue="zcp-nova", exclusive=True)
            channel.queue_bind(exchange='nova', queue="zcp-nova",
                               routing_key='notifications.#')
            channel.queue_bind(exchange='nova', queue="zcp-nova",
                               routing_key='compute.#')
            channel.basic_consume(self.nova_callback,
                                  queue="zcp-nova",
                                  no_ack=True)
            channel.start_consuming()
        except Exception as e:
            LOG.error("Fail to consume messages from nova: %s" % e)
            # Make sure consume messages normally
            self.mqconnc()

    def nova_callback(self, ch, method, properties, body):
        """
        Method used by method nova_amq() to filter messages by type of message.

        :param ch: refers to the head of the protocol
        :param method: refers to the method used in callback
        :param properties: refers to the proprieties of the message
        :param body: refers to the message transmitted
        """
        payload = json.loads(body)

        try:
            project_name = payload.get('_context_project_name')
            project_id = payload.get('_context_project_id')
            if not (project_id or project_name):
                LOG.debug("Drop a notification message,"
                          "project_name or project_id is not found.")
                return
            type_of_message = payload['event_type']

            if type_of_message == 'compute.instance.create.end':
                instance_id = payload['payload']['instance_id']
                instance_name = payload['payload']['hostname']
                self.zabbix_handler.create_host(instance_name,
                                                instance_id,
                                                project_name,
                                                project_id)
                LOG.info("Creating a host: %s(%s) in Zabbix Server"
                         % (instance_id, instance_name))
            elif type_of_message == 'compute.instance.delete.end':
                host = payload['payload']['instance_id']
                host_id = self.zabbix_handler.find_host_id(host)
                self.zabbix_handler.delete_host(host_id)
                LOG.info("Deleting a host: %s in Zabbix Server"
                         % host_id)
            else:
                # TO DO
                # Maybe more event types will be supported
                pass
        except Exception as ex:
            LOG.error('Failed to handler nova event: %(event_type)s '
                      'error message: %(msg)s.'
                      % {'event_type': payload['event_type'], 'msg': ex})



#    Copyright  2017 EasyStack, Inc
#    Author: Branty(jun.wang@easystack.cn)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
from logging import handlers
import os

from eszcp.common.conf import Conf


cfg_file = Conf()
root_level = cfg_file.read_option('log', 'root_level')
console_level = cfg_file.read_option('log', 'consolo_level')
log_level = cfg_file.read_option('log', 'log_level')
log_dir = cfg_file.read_option('log', 'log_dir')
log_file = cfg_file.read_option('log', 'log_file')
maxbytes = cfg_file.read_option('log', 'maxbytes')
backupcount = cfg_file.read_option('log', 'backupcount')
log_default_format = cfg_file.read_option('log',
                                          'log_default_format',
                                          raw=True)


class ContextAdapt(object):
    def __init__(self, logger):
        self.logger = logger

    def debug(self, msg=None):
        if msg:
            self.logger.debug(msg)

    def DEBUG(self, msg=None):
        self.debug(msg)

    def info(self, msg=None):
        if msg:
            self.logger.info(msg)

    def INFO(self, msg=None):
        self.info(msg)

    def warning(self, msg=None):
        if msg:
            self.logger.warning(msg)

    def WARING(self, msg=None):
        self.warning(msg)

    def error(self, msg=None):
        if msg:
            self.logger.error(msg)

    def ERROR(self, msg=None):
        self.error(msg)

    def critical(self, msg=None):
        if msg:
            self.msg.critical(msg)

    def CRITICAL(self, msg=None):
        self.critical(msg)

_loggers = {}
VALID_LEVELS = [
    'CRITICAL', 'critical',
    'FATAL', 'fatal',
    'ERROR', 'error',
    'WARNING', 'warning',
    'WARN', 'warn',
    'INFO', 'info',
    'DEBUG', 'debug',
    'NOTSET', 'noset'
]


def logger(name=None):
    if name not in _loggers:
        _loggers[name] = ContextAdapt(logging.getLogger(name))
    return _loggers[name]


def prepare_init():
    if not os.path.exists(log_dir):
        msg = "No such directory: %s" % log_dir
        raise IOError(msg)
    if not root_level or root_level not in VALID_LEVELS:
        msg = "Invalid log level of root_level: %s" % root_level
        raise ValueError(msg)
    if not console_level or console_level not in VALID_LEVELS:
        msg = "Invalid log level of consolo_level: %s" \
                % console_level
        raise ValueError(msg)
    if not log_level or log_level not in VALID_LEVELS:
        msg = "Invalid log level of log_level: %s" % log_level
        raise ValueError(msg)


def initlog():
    prepare_init()
    logfile = log_dir + log_file \
        if log_dir.endswith("/") else \
        log_dir + '/' + log_file
    # set root logger
    rootlogger = logger().logger
    rootlogger.setLevel(root_level.upper())
    logfmt = logging.Formatter(log_default_format)
    loghdl = handlers.RotatingFileHandler(filename=logfile,
                                          maxBytes=int(maxbytes),
                                          mode='a',
                                          backupCount=int(backupcount))
    consolohdl = logging.StreamHandler()
    # set log_file logger
    loghdl.setFormatter(logfmt)
    loghdl.setLevel(log_level.upper())

    # set log_file IOStream logger
    consolohdl.setLevel(console_level.upper())
    consolohdl.setFormatter(logfmt)
    # record log message in log_file or print log message with standard output
    rootlogger.addHandler(loghdl)
    rootlogger.addHandler(consolohdl)



#    Copyright  2017 EasyStack, Inc
#    Authors: Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import time

from eszcp.common import conf
from eszcp.common import log
from eszcp import exceptions

LOG = log.logger(__name__)

SUPPORTED_HANDLERS = {
    'ceilometer': 'eszcp.task.polling.ceilometer_handler',
    'mongodb': 'eszcp.task.polling.mongodb_handler'
    }


class Handler(object):
    """Base class for polling handler.
    """

    def __init__(self, conf, zabbix_hdl):
        self.conf = conf
        self.zabbix_hdl = zabbix_hdl

    def interval_run(self, func=None):
        """
        :param func: loop execute function
        """
        LOG.info("********* Polling Ceilometer Metrics Into Zabbix **********")
        while True:
            self.run()
            time.sleep(self.polling_interval)

    def run(self):
        hosts, hosts_map = self.zabbix_hdl.get_hosts(filter_no_proxy=True)
        # Refresh zabbix hostgroups and proxies if needed
        self.zabbix_hdl.check_host_groups()
        self.zabbix_hdl.check_proxies()
        self.polling_metrics(hosts, hosts_map)

    def polling_metrics(self, instance_id, proxy_name):
        """
        :param instance_id: nova instance uuid
        :param proxy_name: zabbix proxy_name
        """
        raise exceptions.NotImplementedError


class HandlerAdapter(object):
    """
    Handler factory
    """
    @staticmethod
    def get_handler(conf, *args):
        """
        :params args: handler requiremnets params
         The following parameters must be specified in 'args' by order:
         param 1: zcp configs
         param 2: a keystone client instance
         param 3: a nova client instance
         param 4: a zabbix handler instance
        """
        cfg = conf if conf else conf.Conf()
        polling_handler = cfg.read_option('zcp_configs',
                                          'polling_handler',
                                          'mongodb')
        if polling_handler not in SUPPORTED_HANDLERS:
            LOG.error('%s not in supported handlers %s'
                      % (polling_handler, SUPPORTED_HANDLERS.keys()))
            raise exceptions.NotImplementedError
        try:
            module = __import__(SUPPORTED_HANDLERS.get(polling_handler),
                                fromlist=['eszcp'])
            return module.get_handler(conf, *args)
        except ImportError or ValueError as e:
            LOG.error('Module %s not found in  Python package %s'
                      % (SUPPORTED_HANDLERS.get('polling_handler'),
                         'eszcp'))
            raise
        except AttributeError or TypeError as e:
            msg = getattr(e, 'msg', '') or getattr(e, 'message', '')
            LOG.error(msg)
            raise
        except Exception as e:
            raise



#!/usr/bin/env python
from setuptools import find_packages
from setuptools import setup

setup(
    version="1.0.1",
    author="hanxi.liu",
    author_email="apolloliuhx@gmail.com",
    packages=find_packages(),
    scripts=['bin/eszcp-polling'],
    description="A Timer task for polling ceilometer metrics into zabbix"
)



#    Copyright  2017 EasyStack, Inc
#    Authors: Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
from ceilometerclient.v2 import client as clm_clientv20

from eszcp.common import log
from eszcp.common import conf

CONF = conf.Conf()
LOG = log.logger(__name__)


def logged(func):

    @functools.wraps(func)
    def with_logging(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception, ex:
            msg = getattr(ex, 'message', None) or \
                  getattr(ex, 'msg', '')
            LOG.error(msg)
            raise
    return with_logging


class Client(object):
    def __init__(self):
        v3_kwargs = {
                "username": CONF.read_option('keystone_authtoken',
                                             'username'),
                "password": CONF.read_option('keystone_authtoken',
                                             'password'),
                "project_name": CONF.read_option(
                                             'keystone_authtoken',
                                             'project_name'),
                "user_domain_name": CONF.read_option(
                                             'keystone_authtoken',
                                             'user_domain_name'),
                "project_domain_name": CONF.read_option(
                                             'keystone_authtoken',
                                             'project_domain_name'),
                "auth_url": CONF.read_option('keystone_authtoken',
                                             'auth_url'),
                "region_name": CONF.read_option(
                                             'keystone_authtoken',
                                             'region_name'),
        }
        self.clm_client = clm_clientv20.Client('', **v3_kwargs)

    @logged
    def list_resources(self, q=None, links=None, limit=None):
        if not isinstance(q, list):
            # TO DO
            # add something warning
            raise
        return self.clm_client.resources.list(q=q,
                                              links=links,
                                              limit=limit)

    @logged
    def statistics(self, meter_name, q=None, limit=None):
        if not isinstance(q, list):
            LOG.error("Invalid query param q: %s,q must be a list" % q)
            raise
        return self.clm_client.statistics.list(meter_name,
                                               q=q,
                                               limit=limit
                                               )



#    Copyright  2017 EasyStack, Inc
#    Authors: Claudio Marques,
#             David Palma,
#             Luis Cordeiro,
#             Branty <jun.wang@easystack.cn>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Class for reading the configuration file

Uses the ConfigParser lib to return the values present in the config file
"""

import ConfigParser
from ConfigParser import NoOptionError
from ConfigParser import NoSectionError
import os
import sys


FIND_DIRS = [
            os.path.abspath(os.path.join(
                            os.path.dirname(__file__),
                            "..", "etc/proxy.conf")
                            ),
            "/etc/zcp/proxy.conf",
            ]

"""
The basic singleton pattern

Use __new__ when you need to control the creation of a new instance.

Use __init__ when you need to control initialization of a new instance.

How to use them,view  the following link:
https://mail.python.org/pipermail/tutor/2008-April/061426.html

When cls._instance is None, the class of Singleton is not  instantiated,
instantiate this class and return.

When cls._instance in not None, return the instance directly.

Talk  is too cheap,show you the codes:

    class Singleton(object):
        def __new__(cls, *args,**kwargs):
            if not hasattr(cls,'_instance'):
               cls._instance = super(Singleton,cls).__new__(cls,
                                                            *args,
                                                            **kwargs)
            return  cls._instance

    class Myclass(Singleton):
        a = 1
    one = Myclass()
    two = Myclass()
    # we can compare one with two, id(), == ,is
    two.a = 3
    print one.a  # output is : 3
    print id(one) == id(two) # outout is : True

"""


def singleton(cls, *args, **kwags):
    instance = {}

    def _singleton():
        if cls not in instance:
            instance[cls] = cls(*args, **kwags)
        return instance[cls]
    return _singleton


@singleton
class Conf:
    """
    parse ZabbixCeilometer Proxy conf , singleton partten

    The way of  decorator for singleton pattern is more pythonic

    """

    config = None

    def __init__(self, conf_file=None):

        """
        Method to read from conf file specific options

        :param file: the zcp configuration file
        """
        zcp_conf = None
        if conf_file and os.path.exists(conf_file):
            zcp_conf = conf_file
        if not zcp_conf:
            for f in FIND_DIRS:
                if os.path.exists(f):
                    zcp_conf = f
                    break
        if not zcp_conf:
            print "Can't find zabbixceilometer-proxy configurate file"
            sys.exit(1)
        self.config = ConfigParser.SafeConfigParser()
        self.config.readfp(open(zcp_conf))

    def read_option(self, group, name, default=None, raw=False):
        """
        :return:
        """
        value = None
        try:
            value = self.config.get(group, name, raw=raw)
        except NoOptionError or NoSectionError:
            if default is not None:
                return default
            else:
                raise
        except Exception:
            raise
        return value
